#include "ko.h"

/* static variable to store the given input parameters */
static InputArgs input;

/* convert a number of given base into a binary code */
unsigned char *convertToBinary(int num, int base)
{
	unsigned char *binaryCode;
	int curr, idx_2;

	binaryCode = malloc(base * sizeof(unsigned char));
	curr = num;
	/* create the binary code */
	for (idx_2 = base; idx_2 >0; idx_2--)
	{
		binaryCode[idx_2-1] = curr % 2;
		curr /= 2;
	}
	return binaryCode;
}

/* the algorithm of Jong-Gook et al. */
unsigned char *algoKo (char const *filename, int cellWidth, int cellHeight, int groupSize, int post)
{
	unsigned char *grayMeans;
	unsigned char *irisCode;
	short int     *cumSums;

	Image *img;

	/* load the given image */
	img = load_tiff_luminance(filename);

	/* initialize the arguments struct */
	initializeInputArgs(img, cellWidth, cellHeight, groupSize);

	/* slit the image to the apropriate dimension */
	img = slitImage(img);

	/* calculate the grayscale means of the image for given cell dimension */
	grayMeans = calculateMeans(img);

	/*printMeans(grayMeans);*/

	/* calculate the cumulative sums of these values */
	cumSums = calculateCumSums(grayMeans);

	/*printGroups(cumSums);*/

	/* finally get the appropriate iris code */
	irisCode = calculateIrisCode(cumSums, post);

	/*printIrisCode(irisCode);*/

	free(img);
	free(grayMeans);
	free(cumSums);

	return irisCode;
}

/* initialze a 'InputArgs' struct with the given input data */
void initializeInputArgs (Image *img, int cellWidth, int cellHeight, int groupSize)
{
	input.cellWidth  = cellWidth;
	input.cellHeight = cellHeight;
	input.imgWidth   = img->w / 2;
	input.imgHeight  = img->h;
	input.groupSize  = groupSize;

	/* the number of horizontal and vertical cells */
	input.horCells = (int)(input.imgWidth/(input.cellWidth * input.groupSize)) * input.groupSize;
	input.verCells = (int)(input.imgHeight/(input.cellHeight * input.groupSize)) * input.groupSize;

	/* the dimension of the resulting image */
	input.imgDim = input.horCells * input.verCells;

}

/* calculate the iris code of given cumulative sums */
unsigned char * calculateIrisCode (short int *cumSums, int post)
{
	int idx_1, idx_2;
	int max, maxPos;
	int min, minPos;
	int idy;
	int up = 0;
	int down = 0;

	unsigned char *iriscode;

	iriscode = malloc(input.imgDim * sizeof(unsigned char) * 4);

	for (idx_1 = 0; idx_1 < input.horCells * input.verCells * 2; idx_1+=input.groupSize)
	{
		/* the initial values and positions */
		max = cumSums[idx_1];
		min = cumSums[idx_1];
		maxPos = 0;
		minPos = 0;

		/* fist calculate the position of the maximal and minimal value */
		for (idx_2 = idx_1; idx_2 < idx_1 + input.groupSize; idx_2++)
		{
			if ( cumSums[idx_2] < min )
			{
				min    = cumSums[idx_2];
				minPos = idx_2 - (idx_1);
			}
		}

		for (idx_2 = idx_1; idx_2 < idx_1 + input.groupSize; idx_2++)
		{
			if ( cumSums[idx_2] >= max )
			{
				max    = cumSums[idx_2];
				maxPos = idx_2 - (idx_1);
			}
		}

		/* check wether there is a up- of a downward slope */
		if (minPos <= maxPos)
		{
			for (idy = minPos + idx_1; idy < maxPos + idx_1 -1; idy++)
			{
				if (cumSums[idy] > cumSums[idy+1])
				{
					up = 0;
					break;
				}
				else
				{
					up = 1;
				}
			}
		}
		else if (minPos > maxPos)
		{
			for (idy = maxPos + idx_1; idy < minPos + idx_1 -1; idy++)
			{
				if (cumSums[idy] < cumSums[idy+1])
				{
					down = 0;
					break;
				}
				else
				{
					down = 1;
				}
			}
		}

		
		/*printf("min: %d, max: %d\n", minPos, maxPos); */
		if (abs(minPos-maxPos)<2 || up == down)
		{
			up = 0;
			down = 0;
		}

		/* now generate the iris code */
		for (idx_2 = idx_1; idx_2 < idx_1 + input.groupSize; idx_2++)
		{
			if ((up == 0)&&(down == 0))
			{
				iriscode[idx_2] = 0;
				iriscode[idx_2 + input.horCells * input.verCells * 2] = 0;
			}
			/* check wether values are outside the borders */
			else if ( (idx_2 - idx_1) < MIN(minPos, maxPos) || (idx_2 - idx_1) > MAX(minPos, maxPos) )
			{
				iriscode[idx_2] = 0;
				iriscode[idx_2 + input.horCells * input.verCells * 2] = 0;
			}
			
			else
			{
				/* check wether it is the last value or the right border  and take the previous value are reference */
				if ((idx_2 - idx_1) == MAX(minPos, maxPos) )
				{
					/* greater value: assign '1' */
					if ( (cumSums[idx_2] >= cumSums[idx_2-1]) && (up==1))
					{
						iriscode[idx_2] = 1;
						iriscode[idx_2 + input.horCells * input.verCells * 2] = 0;
					}
					/* smaller value: assign '2' */
					else if ( (cumSums[idx_2] < cumSums[idx_2-1]) && (down==1))
					{
						iriscode[idx_2] = 0;
						iriscode[idx_2 + input.horCells * input.verCells * 2] = 1;
					}
					else
					{
						iriscode[idx_2] = 0;
						iriscode[idx_2 + input.horCells * input.verCells * 2] = 0;
					}
					
				}
				else
				{
					/* the common case, where just the next value is checked to apply the code */
					/* greater value: assign '1' */
					/* same value: assign this value */
					if ( cumSums[idx_2] == cumSums[idx_2+1] )
					{
						if (up == 1)
						{
							iriscode[idx_2] = 1;
							iriscode[idx_2 + input.horCells * input.verCells * 2] = 0;
						}
						else if (down == 1)
						{
							iriscode[idx_2] = 0;
							iriscode[idx_2 + input.horCells * input.verCells * 2] = 1;
						}
						else
						{
							iriscode[idx_2] = 0;
							iriscode[idx_2 + input.horCells * input.verCells * 2] = 0;
						}
					}
					else if (up==1)
					{
						iriscode[idx_2] = 1;
						iriscode[idx_2 + input.horCells * input.verCells * 2] = 0;
					}
					/* smaller value: assign '2' */
					else if (down==1)
					{
						iriscode[idx_2] = 0;
						iriscode[idx_2 + input.horCells * input.verCells * 2] = 1;
					}
				}
			}
			/*printf("%d", iriscode[idx_2]);*/	
		}
		/*printf("\n");*/
	}
	return iriscode;
}

/* claculate the cumulative sums of all given mean values (short int is necessary) */
short int * calculateCumSums (unsigned char *means)
{
	int idx_1, idx_2, idx_3;
	int mean;
	int start;

	short int *sums;

	sums = malloc(input.imgDim * sizeof(short int) * 2);

	/* first calculate all the horizontal groups */
	start = 0;
	for (idx_1 = 0; idx_1 < input.verCells; idx_1++)
	{
		for (idx_2 = 0; idx_2 < input.horCells; idx_2+=input.groupSize)
		{
			mean = 0;
			/* calculate the mean of all gryscale values */
			for (idx_3 = idx_2; idx_3 < idx_2 + input.groupSize; idx_3++)
			{
				mean += means[idx_3 + idx_1 * input.horCells];
			}
			mean /= input.groupSize;

			/* calculate the cumulative sum for the group */
			sums[start] = means[start] - mean;
			
			for (idx_3 = start+1; idx_3 < start + input.groupSize; idx_3++)
			{
				sums[idx_3] = sums[idx_3-1];
				sums[idx_3] += means[idx_3] - mean;
			}
			start += input.groupSize;
		}
	}

	/* now calculate the vertical groups */
	start = input.horCells * input.verCells;

	for (idx_1 = 0; idx_1 < input.verCells; idx_1+=input.groupSize)
	{
		for (idx_2 = 0; idx_2 < input.horCells; idx_2++)
		{
			mean = 0;
			/* calculate the mean of all gryscale values */
			for (idx_3 = idx_2; idx_3 < idx_2 + input.groupSize; idx_3++)
			{
				mean += means[idx_2 + (idx_1 + idx_3 - idx_2) * input.horCells];
			}
			mean /= input.groupSize;

			/* calculate the cumulative sum for the group */
			sums[start] = means[idx_2 + idx_1 * input.horCells] - mean;
			for (idx_3 = start+1; idx_3 < start + input.groupSize; idx_3++)
			{
				sums[idx_3] = sums[idx_3-1];
				sums[idx_3] += means[idx_2 + (idx_1 + idx_3 - start) * input.horCells] - mean;
			}
			start+=input.groupSize;
		}
	}

	return sums;
}

/* calculate the means of all grayscale values in the cells */
unsigned char *calculateMeans (Image *image)
{
	int idx_1, idx_2, idx_3, idx_4;
	int mean_idx;
	int curr_mean;

	unsigned char *means;

	means = malloc(input.imgDim * sizeof(unsigned char));

	mean_idx = 0;
	/*printf("width: %d, height: %d\n", image->w, image->h);*/

	for (idx_1 = 0; idx_1 < image->h; idx_1+=input.cellHeight)
	{
		for (idx_2 = 0; idx_2 < image->w; idx_2+=input.cellWidth)
		{
			curr_mean = 0;

			/* now process one cell of given width and height */
			for (idx_3 = idx_2; idx_3 < idx_2 + input.cellWidth; idx_3++)
			{
				for (idx_4 = 0; idx_4 < input.cellHeight; idx_4++)
				{
					curr_mean += (int)image->data.b[idx_3 + (idx_4 + idx_1) * image->w];
				}
			}
			curr_mean /= input.cellWidth * input.cellHeight;
			means[mean_idx] = curr_mean;
			mean_idx++;
		}
	}

	return means;
}

/* slit the image so that only half of the image is left over which in the best case
   does not contain any eyelids */
Image * slitImage (Image *image)
{
	int idx_1, idx_2;
	int slitted_w, slitted_h;
	unsigned char *data;

	int hDiff, vDiff;
	int upperBorder, lowerBorder;
	int leftBorder, rightBorder;

	int startPoint, endPoint;

	/* the width and height of the resulting image */
	slitted_w = input.horCells * input.cellWidth;
	slitted_h = input.verCells * input.cellHeight;

	/* the horizontal and vertical difference to the original image */
	hDiff = image->h - slitted_h;
	vDiff = image->w/2 - slitted_w;

	/* the borders of the new image */
	upperBorder = hDiff/2;
	lowerBorder = image->h - hDiff/2;

	/* the borders of the half from 135° to 225° from which 'vDiff' is subtracted */
	leftBorder  = image->w * 3/8 + vDiff/2;
	rightBorder = image->w * 5/8 - vDiff/2;

	/* in case differences are odd stretch borders */
	if ( hDiff % 2 == 1 )
	{
		upperBorder++;
	}

	if ( vDiff % 2 == 1 )
	{
		leftBorder++;
	}

	data  = malloc(slitted_w * slitted_h * sizeof(unsigned char));

	/* process the quater from 45° to 0° (the last eighth of the original image) */
	startPoint = image->w * 7/8;
	endPoint   = image->w;
	for (idx_1 = upperBorder; idx_1 < lowerBorder; idx_1++)
	{
		for (idx_2 = startPoint; idx_2 < endPoint; idx_2++)
		{
			data[idx_2 - startPoint + (idx_1 - upperBorder) * slitted_w] = 
				image->data.b[idx_2 + idx_1 * image->w];
		}
	}

	/* process the quater from 0° to 315° (the first eighth of the original image) */
	startPoint = 0;
	endPoint   = image->w * 1/8;
	for (idx_1 = upperBorder; idx_1 < lowerBorder; idx_1++)
	{
		for (idx_2 = startPoint; idx_2 < endPoint; idx_2++)
		{
			data[idx_2 + endPoint + (idx_1 - upperBorder) * slitted_w] = 
				image->data.b[idx_2 + idx_1 * image->w];
		}
	}

	/* process the  half from 135° to 225° (the forth and fifth eighth of the original image) */
	for (idx_1 = upperBorder; idx_1 < lowerBorder; idx_1++)
	{
		for (idx_2 = leftBorder; idx_2 < rightBorder; idx_2++)
		{
			data[idx_2 - leftBorder + image->w/4 + (idx_1 - upperBorder) * slitted_w] = 
				image->data.b[idx_2 + idx_1 * image->w];
		}
	}

	/*save_tiff_luminance("result.tiff", image_new(data, slitted_w, slitted_h, image->type));*/

	/* return the new image */
	return  image_new(data, slitted_w, slitted_h, image->type);
}

/* return the dimension of the result */
int getImgDim()
{
	return input.imgDim;
}

/* write the result into a '.code' */
void createCodeFile(char *filename, unsigned char *irisCode)
{
	int idx_1, idx_2;
	unsigned char j[1];
	FILE *outFile;
	char *p;

	p = strstr (filename, "_finaltexture.tiff");
  	strncpy (p, "\0", 1);

	outFile = fopen(strcat(filename, "_ko.code"), "w");

	if ( outFile == NULL )
	{
		fprintf(stderr,"Can not open output file\n");
		exit(1);
	}
	
	for (idx_1 = 0; idx_1 < input.imgDim * 4; idx_1+= 8)
	{
		j[0] = 0;
		for (idx_2 = 0; idx_2 < 8; idx_2++)
		{	
			j[0] += ((int)irisCode[idx_1+7-idx_2] * (int)(pow(2, idx_2)));
		}
		fwrite(j, sizeof(char), 1, outFile);
	}
	
	printf("%d bits extracted\n", idx_1);
	printf("output written to %s\n", filename);
	fclose(outFile);
}

/* print the result */
void printIrisCode(unsigned char *irisCode)
{
	int idx;
	for (idx = 0; idx < input.imgDim * 4; idx++)
	{
		printf("%d", (int)irisCode[idx]);
	}
	printf("\n");
}

/* print the calculated means */
void printMeans(unsigned char *means)
{
	int idx;
	printf("Means:");
	for (idx = 0; idx < input.imgDim; idx++)
	{
		printf("%d: %d; ",idx, (int)means[idx]);
	}
	printf("\n");
}

/* print the groups of the calculated sums */
void printGroups(short int *sums)
{
	int idx;

	/* the horizontal groups */
	printf("H_Groups:");
	for(idx = 0; idx < input.imgDim; idx++)
	{
		if ( idx % input.groupSize == 0 )
		{
			printf("\n");
		}
		printf("%d; ", (int)sums[idx]);
	}
	printf("\n");

	/* the vertical groups */
	printf("V_Groups:");
	for(idx = input.imgDim; idx< input.imgDim * 2; idx++)
	{
		if ( idx % input.groupSize == 0 )
		{
			printf("\n");
		}
		printf("%d; ", (int)sums[idx]);
	}
	printf("\n");
}

int getFileSize(char *filename)
{
	int filesize;
	struct stat buf;

	if(stat(filename, &buf) < 0) {
		exit(-1);
	}
	filesize = buf.st_size;

	return filesize;
}

/* convert a file into a binary stream */
unsigned char *binaryReadFile(char *filename, int filesize)
{
	unsigned char *binaryCode;
	int curr, idx_1, idx_2;
	unsigned char buffer[1];
	FILE *file;

	binaryCode = malloc(filesize * sizeof(unsigned char) * 8);

	file = fopen(filename, "r");
	
	for (idx_1 = 0; idx_1 < filesize; idx_1++)
	{
		curr =0;
		buffer[0] = 0;
		fread(buffer, 1, 1, file);

		curr = (int)buffer[0];

		/* create the binary code */
		for (idx_2 = 0; idx_2 < 8; idx_2++)
		{
			binaryCode[idx_1 * 8 + 7 - idx_2] = curr % 2;
			curr /= 2;
		}
	}
	
	fclose(file);
	return binaryCode;
}
